#include "ObjectAxes.h"
#include <math.h>
#define PI 3.1415926535

ObjectAxes::ObjectAxes(int x,int y,int w,int h,const char *l) 
: Fl_Gl_Window(x,y,w,h,l) {

	// initialize the model parameters 

	theta = 0.0;
	phi = 45;
	psi = 90;

	h1 = 2;			// base
	h2 = 3;			// lower arm
	h3 = 2;			// upper arm

	// Initialize the the vertex arrays for the two arm boxes.

	GLfloat box[8][3] = {
		{-1.0,0,-1.0}, {-1.0,0,1.0}, {-1.0,1,-1.0}, {-1.0,1,1.0},
		{ 1.0,0,-1.0}, { 1.0,0,1.0}, { 1.0,1,-1.0}, { 1.0,1,1.0}
	};

	GLfloat scale = (GLfloat).5;
	for (int i=0; i<8; i++) {				// for each vertex
		boxLower[i][0] = scale*box[i][0];	// scale x dimension
		boxLower[i][1] = h2 * box[i][1];	// set y to correct height
		boxLower[i][2] = scale*box[i][2];	// scale z dimension
	}

	scale = (GLfloat).4;
	for (int i=0; i<8; i++) {				// for each vertex
		boxUpper[i][0] = scale*box[i][0];	// scale x dimension
		boxUpper[i][1] = h3 * box[i][1];	// set y to correct height
		boxUpper[i][2] = scale*box[i][2];	// scale z dimension
	}
 
	// Identify which vertices to use for each face.  They are in
	// counter clockwise order, looking from the front side of the face.

	int orderInit[6][4] = {
		{0,1,3,2}, {4,6,7,5},	// x=0, x=1
		{0,4,5,1}, {2,3,7,6},	// y=0, y=1
		{0,2,6,4}, {1,5,7,3}
	};
	for (int i=0; i<6; i++)			// for each face
		for (int j=0; j<4; j++)		// for each vertex subscript
			order[i][j] = orderInit[i][j];

	// Identify the shading normal for each face.

	GLfloat normalInit[6][3] = {
		{-1,0,0}, {1,0,0},
		{0,-1,0}, {0,1,0},
		{0,0,-1}, {0,0,1}
	};
	for (int i=0; i<6; i++)			// for each face
		for (int j=0; j<3; j++)		// for each dimension of the normal
			normal[i][j] = normalInit[i][j];

	// Build the quadric object that will be used for the base.

	qobj = gluNewQuadric();
	//	gluQuadricCallback(qobj,GLU_ERROR,qerrCallback);
	gluQuadricDrawStyle(qobj,GLU_FILL);
	gluQuadricNormals(qobj,GLU_SMOOTH);

	// Initialize the transformation matrices to the identity
	// These are 4x4 matrices in column major order

	for (int i=0; i<16; i++) {
		m1[i] = m2[i] = m3[i] = ((i % 5) == 0) ? (GLfloat)1.0 : (GLfloat)0.0;
		mRtheta[i] = mTh1[i] = mRphi[i] = mTh2[i] = mRpsi[i] = m1[i];
	}

	mTh1[13] = h1;
	mTh2[13] = h2;
};
/**
* Set theta, base rotation about y axis.
*/
void ObjectAxes::setTheta(float angle){
	theta=angle;
	updateMatrices();
};
/**
* Set phi, lower arm rotation about z axis.
*/
void ObjectAxes::setPhi(float angle){
	phi=angle;
	updateMatrices();
};
/**
  * Set psi, upper arm rotation about z axis.
  */
void ObjectAxes::setPsi(float angle){
	psi=angle;
	updateMatrices();
};
/**
 * Draw all the objects in this axes.
 */
void ObjectAxes::draw(void) {
	
	GLfloat light_position[] = {0.0,0.0,1.0,0.0};
	GLfloat gray_light[] = {0.2,0.2,0.2,1.0};
	GLfloat white_light[] = {1.0,1.0,1.0,1.0};
	GLfloat no_light[] = {0.0,0.0,0.0,1.0};

	if (!valid()) {
		glViewport(0,0,w(),h());

		// perspective projection

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity(); 
		glFrustum(-1.0,1.0,-1.0,1.0,1.0,20.0);
		gluLookAt(0.0,4.0,6.0,0.0,2.0,0.0,0.0,1.0,0.0);

		// modeling and lighting

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity(); 

		glLightfv(GL_LIGHT0,GL_POSITION, light_position);
		glLightfv(GL_LIGHT0,GL_AMBIENT, gray_light);
		glLightfv(GL_LIGHT0,GL_DIFFUSE, white_light);
		glLightfv(GL_LIGHT0, GL_SPECULAR, no_light);

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_DEPTH_TEST);
		glClearColor(1.0,1.0,1.0,1.0);
		glShadeModel(GL_SMOOTH);
	}

	// Clear the scene

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// draw it again

	glPushMatrix(); 
	glMultMatrixf(mRtheta);		// glRotatef(theta,0,1,0); 
	base();
	glMultMatrixf(mTh1);		// glTranslatef(0,h1,0);
	glMultMatrixf(mRphi);		// glRotatef(phi,0,0,1); 
	lower_arm();
	glMultMatrixf(mTh2);		// glTranslatef(0,h2,0);
	glMultMatrixf(mRpsi);		// glRotatef(psi,0,0,1); 
	upper_arm();
	glPopMatrix();

	// check for any GL errors

	err = glGetError();
	if (err != GL_NO_ERROR) {
		int e = err;
	}
};
/**
 * Draw the base cylinder.  Rotate the axes to do it because
 * gluCylinder draws along the +z axis.
 */
void ObjectAxes::base(void) {
	GLfloat radius = 1;
	GLint slices = 16;
	GLfloat mat_dark_red[] = {.4, 0.0, 0.0, 1.0};
	glMaterialfv(GL_FRONT,GL_AMBIENT_AND_DIFFUSE,mat_dark_red);
	glPushMatrix();
	glRotatef(-90,1,0,0);
	gluCylinder(qobj,radius,radius,h1,slices,1);
	glTranslatef(0,0,h1);
	gluDisk(qobj,0,radius,slices,1);
	glPopMatrix();
};
/**
 * Draw the lower arm.
 */
void ObjectAxes::lower_arm(void) {
	GLfloat mat_dark_green[] = {0.0, .4, 0.0, 1.0};
	glMaterialfv(GL_FRONT,GL_AMBIENT_AND_DIFFUSE,mat_dark_green);
	glBegin(GL_QUADS);
	for (int i=0; i<6; i++) {			// each face
		glNormal3fv(normal[i]);
		for (int j=0; j<4; j++) {		// each vertex
			glVertex3fv(boxLower[ order[i][j] ]);
		}
	}
	glEnd();
};
/**
 * Draw the upper arm.
 */
void ObjectAxes::upper_arm(void) {
	GLfloat mat_dark_blue[] = {0.0, 0.0, 0.4, 1.0};
	glMaterialfv(GL_FRONT,GL_AMBIENT_AND_DIFFUSE,mat_dark_blue);
	glBegin(GL_QUADS);
	for (int i=0; i<6; i++) {			// each face
		glNormal3fv(normal[i]);
		for (int j=0; j<4; j++) {		// each vertex
			glVertex3fv(boxUpper[ order[i][j] ]);
		}
	}
	glEnd();
};
/**
 * Update all the transformation matrices.
 * These matrices are 4x4 in row major order.
 */
void ObjectAxes::updateMatrices(void) {

	// base = Rtheta

	GLfloat r = theta * PI / 180.;
	mRtheta[0] = cos(r);
	mRtheta[2] = -sin(r);
	mRtheta[8] = sin(r);
	mRtheta[10] = cos(r);

	for (int i=0; i<16; i++) 
		m1[i] = mRtheta[i];

	// lower arm = Rtheta*Th1*Rphi

	r = phi * PI / 180.;
	mRphi[0] = cos(r);
	mRphi[1] = sin(r);
	mRphi[4] = -sin(r);
	mRphi[5] = cos(r);

	// upper arm = Rtheta*Th1*Rphi*Th2*Rphi

	for (int i=0; i<16; i++) 
		m2[i] = mRtheta[i];
	multiply(m2,mTh1);
	multiply(m2,mRphi);

	r = psi * PI / 180.;
	mRpsi[0] = cos(r);
	mRpsi[1] = sin(r);
	mRpsi[4] = -sin(r);
	mRpsi[5] = cos(r);

	for (int i=0; i<16; i++) 
		m3[i] = mRtheta[i];
	multiply(m3,mTh1);
	multiply(m3,mRphi);
	multiply(m3,mTh2);
	multiply(m3,mRpsi);
};
/**
 * Multiply two 4x4 matrices A and B and return the result in A.
 * These matrices are 4x4 in column major order.
 */
void ObjectAxes::multiply(GLfloat *A,GLfloat *B) {
	int s[4][4] = {
		{0,4,8,12},
		{1,5,9,13},
		{2,6,10,14},
		{3,7,11,15}
	};
	GLfloat C[16];
	for (int i=0; i<4; i++) {
		for (int j=0; j<4; j++) {
			C[s[i][j]] = 
				A[s[i][0]]*B[s[0][j]] + 
				A[s[i][1]]*B[s[1][j]] + 
				A[s[i][2]]*B[s[2][j]] + 
				A[s[i][3]]*B[s[3][j]];
		}
	}
	for (int i=0; i<16; i++) 
		A[i] = C[i];
}


